"""Helper functions"""
import re
import time
from typing import Any, Iterable

from netports.static import BRIEF_ALL_I, BRIEF_ALL_S, SPLITTER
from netports.types_ import LAny, LStr, StrInt, IStrInt, LInt, T2Str, T3Str, T4Str


# =============================== str ================================


def findall1(pattern: str, string: str, flags=0) -> str:
    """Parses 1st item of re.findall(). If nothing is found, returns an empty string.
    Group with parentheses in pattern is required
    ::
        :return: Interested substring
        :example:
            pattern: "a(b)cde"
            string: "abcde"
            return: "b"
    """
    result = (re.findall(pattern=pattern, string=string, flags=flags) or [""])[0]
    if isinstance(result, str):
        return result
    if isinstance(result, tuple):
        return result[0]
    return ""


def findall2(pattern: str, string: str, flags=0) -> T2Str:
    """Parses 2 items of re.findall(). If nothing is found, returns 2 empty strings.
    Group with parentheses in pattern is required
    ::
        :return: Two interested substrings
        :example:
            pattern: "a(b)(c)de"
            string: "abcde"
            return: "b", "c"
    """
    result = (re.findall(pattern=pattern, string=string, flags=flags) or [("", "")])[0]
    if isinstance(result, tuple) and len(result) >= 2:
        return result[0], result[1]
    return "", ""


def findall3(pattern: str, string: str, flags=0) -> T3Str:
    """Parses 3 items of re.findall(). If nothing is found, returns 3 empty strings.
    Group with parentheses in pattern is required
    ::
        :return: Three interested substrings
        :example:
            pattern: "a(b)(c)(d)e"
            string: "abcde"
            return: "b", "c", "d"
    """
    result = (re.findall(pattern=pattern, string=string, flags=flags) or [("", "", "")])[0]
    if isinstance(result, tuple) and len(result) >= 3:
        return result[0], result[1], result[2]
    return "", "", ""


def findall4(pattern: str, string: str, flags=0) -> T4Str:
    """Parses 4 items of re.findall(). If nothing is found, returns 4 empty strings
    ::
        :param pattern: Regex pattern, where 4 groups with parentheses in pattern are required
        :param string: String where need to find pattern
        :param flags: findall flags
        :return: Three interested substrings
        :example:
            pattern = "a(b)(c)(d)(e)f"
            string = "abcdef"
            return: "b", "c", "d", "e"
    """
    result = (re.findall(pattern=pattern, string=string, flags=flags) or [("", "", "", "")])[0]
    if isinstance(result, tuple) and len(result) >= 4:
        return result[0], result[1], result[2], result[3]
    return "", "", "", ""


def repr_params(*args, **kwargs) -> str:
    """Makes params for __repr__() method"""
    args_ = ", ".join([f"{v!r}" for v in args if v])
    kwargs_ = ", ".join([f"{k}={v!r}" for k, v in kwargs.items() if v])
    params = [s for s in (args_, kwargs_) if s]
    return ", ".join(params)


def join(items: LAny) -> str:
    """Joins items by "," """
    return SPLITTER.join([str(i) for i in items])


# =============================== bool ===============================

def is_all(**kwargs) -> bool:
    """Get param `all` from `kwargs`"""
    return bool(kwargs.get("all"))


def is_brief(**kwargs) -> bool:
    """Get param `verbose` from `kwargs` and reverse the result"""
    return not is_verbose(**kwargs)


def is_brief_in_items(items: Any) -> bool:
    """Checks is "-1" or -1 in `items`, used for verbose=False
    ::
        :param items: *str, int, List[str], List[int]*
        :return: True - if "-1" or -1 present in `items`
    """
    if isinstance(items, (str, int)):
        items = [items]
    if not isinstance(items, Iterable):
        raise TypeError(f"{items=} expected {str}")
    for item in items:
        if isinstance(item, str):
            if item == BRIEF_ALL_S:
                return True
            continue
        if isinstance(item, int):
            if item == BRIEF_ALL_I:
                return True
            continue
        raise TypeError(f"{item=} expected {str}")
    return False


def is_verbose(**kwargs) -> bool:
    """Get param `verbose` from `kwargs`"""
    verbose = kwargs.get("verbose")
    if verbose is None:
        return False
    return bool(verbose)


def is_strict(**kwargs) -> bool:
    """Get param `strict` from `kwargs`"""
    strict = kwargs.get("strict")
    if strict is None:
        return True
    return bool(strict)


# =============================== int ================================

def to_int(number: StrInt) -> int:
    """Converts *str* to *int*
    ::
        :param number: *int* or digit as *str*
        :return: *int*
        :raises TypeError: If number is not digit
    """
    if isinstance(number, int):
        return number
    if isinstance(number, str) and number.isdigit():
        return int(number)
    raise TypeError(f"{number=} {int} expected")


def to_lint(numbers: IStrInt) -> LInt:
    """Converts *List[str]* to *List[int]*
    ::
        :param numbers: *List[str]*
        :return: *List[int]*
        :raises TypeError: If number is not digit
    """
    if not isinstance(numbers, (list, set, tuple)):
        raise TypeError(f"{numbers=} {list} expected")
    return [to_int(s) for s in numbers]


# =============================== list ===============================

def lstr(items: Any) -> LStr:
    """Converts any types of items to *List[str]*"""
    if not isinstance(items, (bytes, dict, float, int, list, set, str, tuple)):
        raise TypeError(f"{items=} {list} expected")
    if isinstance(items, (bytes, float, int, str)):
        items = [items]
    items_ = []
    for item in items:
        if isinstance(item, bytes):
            items_.append(item.decode("utf-8"))
        elif isinstance(item, float):
            items_.append(str(int(item)))
        else:
            items_.append(str(item))
    return items_


def no_dupl(items: Iterable) -> list:
    """Remove duplicate items from list"""
    items_ = []
    for item in items:
        if item not in items_:
            items_.append(item)
    return items_


def remove_brief_items(items: Any) -> list:
    """Remove "-1", -1 from items
    ::
        :param items: *str, int, List[str], List[int]*
        :return: Items without "-1", -1
    """
    if isinstance(items, (str, int)):
        items = [items]
    return [i for i in items if i not in (BRIEF_ALL_S, BRIEF_ALL_I)]


def split(items: Any) -> LStr:
    """Splits items *str*, *List[str]* by "," " " """
    results: LStr = []
    for items_s in lstr(items):
        for item_s in items_s.split():
            results.extend(item_s.split(SPLITTER))
    return results


# ============================= wrapper ==============================

def time_spent(func):
    """Wrapper measure function execution time"""

    def wrap(*args, **kwargs):
        """Wrap"""
        started = time.time()
        return_ = func(*args, **kwargs)
        elapsed = time.time() - started
        print("=== {:s}, spent {:.3f}s ===".format(func.__name__, elapsed))
        return return_

    return wrap



"""netports"""

from netports.exceptions import NetportsValueError
from netports.intf import Intf
from netports.intf_gm import IntfGM, intfrange
from netports.intf_map import short_to_long, short_to_short, long_to_short, long_to_long
from netports.ip import IP_NAMES, IP_NUMBERS, iip, sip, ip_pairs
from netports.item import Item
from netports.ports import inumbers, parse_range, snumbers
from netports.range import Range
from netports.swversion import SwVersion
from netports.tcp import stcp, itcp, check_port, check_ports
from netports.vlan import ivlan, svlan

__all__ = [
    "IP_NAMES",
    "IP_NUMBERS",
    "Intf",
    "IntfGM",
    "Item",
    "NetportsValueError",
    "Range",
    "SwVersion",
    "check_port",
    "check_ports",
    "iip",
    "intfrange",
    "inumbers",
    "ip_pairs",
    "itcp",
    "ivlan",
    "long_to_long",
    "long_to_short",
    "parse_range",
    "short_to_long",
    "short_to_short",
    "sip",
    "snumbers",
    "stcp",
    "svlan",
]



"""An object of interface name, that can contain up to 4 indexes.
Sorts the interfaces by indexes (not by alphabetic).
"""
import re
from functools import total_ordering
from typing import List, Optional, Set, Tuple, Union

from netports import intf_map, helpers as h
from netports.exceptions import NetportsValueError
from netports.static import DEVICE_TYPES
from netports.types_ import T3Str, T5Str, LStr, SStr, DStr, LT2Str

SPLITTER = ",./:"


@total_ordering
class Intf:
    """An object of interface name, that can contain up to 4 indexes.
    Sorts the interfaces by indexes (not by alphabetic).
    """

    def __init__(self, line: str = "", **kwargs):
        """Intf
        ::
            :param line: Interface name that can contain up to 4 indexes
            :type line: str
            :param device_type: Netmiko device type (default "")
            :type device_type: str
            :param splitter: Separator of characters between indexes (default ",./:")
            :type splitter: str
        """
        self._device_type = self._init_device_type(**kwargs)
        self._splitter = str(kwargs.get("splitter") or SPLITTER)
        self._line = self._init_line(line)

    def __str__(self):
        return self.line

    def __repr__(self):
        return f"{self.__class__.__name__}({self.line!r})"

    def __hash__(self) -> int:
        return hash((self.id0, self.id1, self.id2, self.id3, self.id4))

    def __eq__(self, other) -> bool:
        """== equality"""
        if self.__class__ == other.__class__:
            if self.__hash__() == other.__hash__():
                return True
        return False

    def __lt__(self, other) -> bool:
        """< less than"""
        if self.__class__ == other.__class__:
            if self.id0 != other.id0:
                return self.id0 < other.id0
            if self.id1 != other.id1:
                return self.id1 < other.id1
            if self.id2 != other.id2:
                return self.id2 < other.id2
            if self.id3 != other.id3:
                return self.id3 < other.id3
            return self.id4 < other.id4
        return False

    # ============================= init =============================

    def _init_line(self, line: str) -> str:
        """Init Interface line"""
        if not isinstance(line, str):
            raise TypeError(f"{line=} {str} expected")
        self._line = line
        self._ids: T5Str = self._get_ids()
        self._dels = self._get_delimiters()

        items = [
            self._ids[0],
            self._ids[1],
            self._dels[0],
            self._ids[2],
            self._dels[1],
            self._ids[3],
            self._dels[2],
            self._ids[4],
        ]
        return "".join(items)

    @staticmethod
    def _init_device_type(**kwargs) -> str:
        """Init Netmiko device type
        ::
            :return: Netmiko device type
            :raise: NetportsValueError if the device type is unknown
        """
        device_type = str(kwargs.get("device_type") or "")
        expected = ["", *DEVICE_TYPES]
        if device_type not in expected:
            raise NetportsValueError(f"{device_type=} {expected=}")
        return device_type

    # =========================== property ===========================

    @property
    def delimiters(self) -> T3Str:
        """Interface all delimiters"""
        return self._dels

    @property
    def id0(self) -> str:
        """Interface name. Line without IDs
        ::
            :example:
                intf = Intf("interface Ethernet1/2/3.4")
                intf.id0 -> "interface Ethernet"
        """
        return self._ids[0]

    @property
    def id1(self) -> int:
        """Interface 1st ID
        ::
            :example:
                intf = Intf("interface Ethernet1/2/3.4")
                intf.id1 -> 1
        """
        return int(self._ids[1]) if self._ids[1] else 0

    @property
    def id2(self) -> int:
        """Interface 2nd ID
        ::
            :example:
                intf = Intf("interface Ethernet1/2/3.4")
                intf.id2 -> 2
        """
        return int(self._ids[2]) if self._ids[2] else 0

    @property
    def id3(self) -> int:
        """Interface 3rd ID
        ::
            :example:
                intf = Intf("interface Ethernet1/2/3.4")
                intf.id3 -> 3
        """
        return int(self._ids[3]) if self._ids[3] else 0

    @property
    def id4(self) -> int:
        """Interface 4th ID
        ::
            :example:
                intf = Intf("interface Ethernet1/2/3.4")
                intf.id4 -> 4
        """
        return int(self._ids[4]) if self._ids[4] else 0

    @property
    def ids(self) -> T5Str:
        """Interface all IDs
        ::
            :example:
                intf = Intf("interface Ethernet1/2/3.4")
                intf.ids -> "interface Ethernet", "1", "2", "3", "4")
        """
        return self._ids

    @property
    def line(self) -> str:
        """Interface line
        ::
            :example:
                intf = Intf("interface Ethernet1/2/3.4")
                intf.line -> "interface Ethernet1/2/3.4"
        """
        return self._line

    @property
    def name(self) -> str:
        """Interface name with IDs
        ::
            :example:
                intf = Intf("interface Ethernet1/2/3.4")
                intf.name -> "Ethernet1/2/3.4"
        """
        return re.sub(r"^interface\s+", "", self.line)

    @property
    def device_type(self) -> str:
        """Netmiko device type"""
        return self._device_type

    @property
    def splitter(self) -> str:
        """Separator of characters between indexes"""
        return self._splitter

    # =========================== methods ============================

    def all_names(self) -> LStr:
        """All variants of names: long, short, upper-case, lover-case
        ::
            :example:
                Intf("Eth1/2").names() -> [
                    "interface Ethernet1/2",
                    "interface ethernet1/2",
                    "interface Eth1/2",
                    "interface eth1/2",
                    "Ethernet1/2",
                    "ethernet1/2",
                    "Eth1/2",
                    "eth1/2",
                ]
        """
        results: SStr = set()
        names_: LStr = [self.line, self.name, self.name_full(), self.name_long(), self.name_short()]
        names_ = h.no_dupl(names_)
        for name_ in names_:
            results.add(name_)
            if not name_.startswith("interface "):
                name_ = f"interface {name_}"
                results.add(name_)

        intf_map_upper: DStr = intf_map.short_to_long(self._device_type)
        intf_map_lower: DStr = intf_map.short_to_long(self._device_type, key_lower=True)
        names: LStr = [self.name, self.name_short()]
        names = h.no_dupl(names)
        for name in names:
            intf_o = Intf(line=name, device_type=self.device_type)
            for id0_short, intf_map_d in [
                (intf_o.id0, intf_map_upper),
                (intf_o.id0.lower(), intf_map_lower),
            ]:
                if id0_long := intf_map_d.get(id0_short) or "":
                    name_long = intf_o.line.replace(id0_short, id0_long, 1)
                    results.add(name_long)
                    if not name_long.startswith("interface "):
                        results.add(f"interface {name_long}")

        results.update([s.lower() for s in results])
        results_: LStr = sorted(results)
        results_.sort(key=len, reverse=True)

        if self.device_type == "hp_procurve":
            if digits := [s for s in results if s.isdigit()]:
                names = [f"interface 1/{s}" for s in digits]
                results_.extend(names)
        return results_

    def last_idx(self) -> int:
        """Index of last ID in interface line
        ::
            :example:
                intf = Intf("interface Ethernet1/2/3.4")
                intf.last_idx() -> 4
        """
        ids = self._ids[1:]
        return len([s for s in ids if s])

    def name_full(self) -> str:
        """Interface long name with IDs and with interface keyword
        ::
            :example:
                intf = Intf("Eth1/2")
                intf.name_full() -> "interface Ethernet1/2"
        """
        name = self.name_long()
        return f"interface {name}"

    def name_long(self) -> str:
        """Interface long name with IDs and without interface keyword
        ::
            :example:
                intf = Intf("Eth1/2")
                intf.name_long() -> "Ethernet1/2"
        """
        id0 = self.id0.lower()
        if id0.startswith("interface "):
            id0 = id0.replace("interface ", "", 1)

        intf_map_short: DStr = intf_map.short_to_long(self._device_type, key_lower=True)
        for short_lower, long_upper in intf_map_short.items():
            if id0 == short_lower:
                id0 = long_upper
                break
        else:
            intf_map_long: DStr = intf_map.long_to_long(self._device_type, key_lower=True)
            for long_lower, long_upper in intf_map_long.items():
                if id0 == long_lower:
                    id0 = long_upper
                    break

        id1 = self.part_after(idx=0)
        name = f"{id0}{id1}"
        return name

    def name_short(self, replace: LT2Str = None) -> str:
        """Interface short name with IDs
        ::
            :param replace: Replace the default short name with the first one
                            that matches in the list of the 'replace' argument.
            :return: Interface short name.
            :example:
                intf = Intf("interface FastEthernet1/2")
                intf.name_short() -> "Fa1/2"
                intf.name_short(replace=[("Fa", "Eth")]) -> "Eth1/2"
        """
        id0 = self.id0.lower()
        if id0.startswith("interface "):
            id0 = id0.replace("interface ", "", 1)

        intf_map_l2s: DStr = intf_map.long_to_short(self._device_type, key_lower=True)
        for long_lower, short_upper in intf_map_l2s.items():
            if id0 == long_lower:
                id0 = short_upper
                break
        else:
            intf_map_s2s: DStr = intf_map.short_to_short(self._device_type, key_lower=True)
            for short_lower, short_upper in intf_map_s2s.items():
                if id0 == short_lower:
                    id0 = short_upper
                    break

        if replace:
            for before, after in replace:
                if id0 == before:
                    id0 = after
                    break

        id1 = self.part_after(idx=0)
        name = f"{id0}{id1}"
        return name

    def part_after(self, idx: int, splitter=True) -> str:
        """Interface part after interested ID
        ::
            :param idx: Interface index
            :param splitter: True - Include splitter from edge, False - Skip splitter from edge
            :return: Part of the interface name after specified interface index
            :example:
                intf = Intf("Ethernet1/2/3.4")
                intf.part_after(2) -> "/3.4"
                intf.part_after(3) -> ".4"
                intf.part_after(2, splitter=False) -> "3.4"
        """
        if idx >= 4:
            return ""

        parts = self._ids[4]
        if idx == 3:
            if splitter:
                return self._dels[2] + parts
            return parts

        parts = self._ids[3] + self._dels[2] + parts
        if idx == 2:
            if splitter:
                return self._dels[1] + parts
            return parts

        parts = self._ids[2] + self._dels[1] + parts
        if idx == 1:
            if splitter:
                return self._dels[0] + parts
            return parts

        parts = self._ids[1] + self._dels[0] + parts
        if idx == 0:
            return parts

        parts = self._ids[0] + parts
        return parts

    def part_before(self, idx: int, splitter=True) -> str:
        """Interface part before interested ID
        ::
            :param idx: Interface index
            :param splitter: True - Include splitter from edge, False - Skip splitter from edge
            :return: Part of the interface name before specified interface index
            :example:
                intf = Intf("Ethernet1/2/3.4")
                intf.part_before(2) -> "Ethernet1/"
                intf.part_before(3) -> "Ethernet1/2/"
                intf.part_before(3, splitter=False) -> "Ethernet1/2"
        """
        if idx <= 0:
            return ""

        parts = self._ids[0]
        if idx == 1:
            return parts

        parts += self._ids[1]
        if idx == 2:
            if splitter:
                return parts + self._dels[0]
            return parts

        parts += self._dels[0] + self._ids[2]
        if idx == 3:
            if splitter:
                return parts + self._dels[1]
            return parts

        parts += self._dels[1] + self._ids[3]
        if idx == 4:
            if splitter:
                return parts + self._dels[2]
            return parts

        return self._line

    # =========================== helpers ============================

    def _get_ids(self) -> T5Str:
        """Splits interface line to name and IDs
        ::
            :example:
                self.line: "interface Ethernet1/2/3.4"
                return: ("interface Ethernet", "1", "2", "3", "4")
        """
        intf = self._line
        for splitter in self._splitter:
            intf = intf.replace(splitter, ",")

        name = r"([a-zA-Z\-\s]+)*"
        id1 = r"(\d+)*"
        id2 = r"(?:,)*(\d+)*"
        pattern = f"{name}{id1}{id2}{id2}{id2}"
        ids: T5Str = (re.findall(pattern, intf) or [("", "", "", "", "")])[0]
        return ids

    def _get_delimiters(self) -> T3Str:
        """Splits interface line to splitters of IDs
        ::
            :example:
                self.line: "interface Ethernet1/2/3.4"
                return: ("/, "/", ".")
        """
        part1 = self._ids[0] + self._ids[1]
        len1 = len(part1)
        delim1 = self._line[len1:len1 + 1]

        part2 = part1 + delim1 + self._ids[2]
        len2 = len(part2)
        delim2 = self._line[len2:len2 + 1]

        part3 = part2 + delim2 + self._ids[3]
        len3 = len(part3)
        delim3 = self._line[len3:len3 + 1]
        return delim1, delim2, delim3


LIntf = List[Intf]
SIntf = Set[Intf]
TIntf = Tuple[Intf]
ULIntf = Optional[Union[str, LStr, Intf, LIntf]]



"""Interface name mapping"""
from netports.types_ import DStr

MAP_OTHER = {  # parent of ios
    "Fo": "FortyGigabitEthernet",
    "Hu": "HundredGigabitEthernet",
    "Two": "TwoGigabitEthernet",
    "V": "Vlan",
    "At": "ATM",
    "EO": "EOBC",
    "FD": "Fddi",
    "MFR": "MFR",
    "Ma": "Management",
    "Mu": "Multilink",
    "Se": "Serial",
    "Twe": "TwentyFiveGigE",
    "Vi": "Virtual-Access",
    "Vt": "Virtual-Template",
}
MAP_CISCO_ASR = {
    "Te": "TenGigE",  # overlapped: ios
    "Hu": "HundredGigE",
    "BE": "Bundle-Ether",
    "Tu": "tunnel-ip",  # low-priority, overlapped: ios, nxos
    "ti": "tunnel-ip",
    "Mg": "MgmtEth",
}
MAP_CISCO_IOS = {  # parent of nxos, asr
    "Eth": "Ethernet",
    "Fa": "FastEthernet",
    "Gi": "GigabitEthernet",
    "Te": "TenGigabitEthernet",  # overlapped: asr
    "Po": "Port-channel",  # overlapped: nxos
    "Lo": "Loopback",  # overlapped: nxos
    "Tu": "Tunnel",  # overlapped: asr
    "Vlan": "Vlan",  # overlapped: h3c
}
MAP_CISCO_NXOS = {
    "Po": "port-channel",  # overlapped: cisco
    "Lo": "loopback",  # overlapped: cisco, asr
    "mgmt": "mgmt",
}
MAP_HP_COMWARE = {  # h3c
    "GE": "GigabitEthernet",
    "Te": "Ten-GigabitEthernet",  # low-priority, overlapped: cisco
    "XGE": "Ten-GigabitEthernet",
    "BAGG": "Bridge-Aggregation",
    "Vlan": "Vlan-interface",  # overlapped: cisco
}
MAP_HP_PROCURVE = {  # hpc
    "Trk": "Trk",
}
ALL_SHORT = sorted({
    *MAP_OTHER,
    *MAP_CISCO_IOS,
    *MAP_CISCO_NXOS,
    *MAP_CISCO_ASR,
    *MAP_HP_COMWARE,
    *MAP_HP_PROCURVE,
})


def long_to_short(device_type: str = "",
                  key_lower: bool = False,
                  value_lower: bool = False) -> DStr:
    """Returns Interfaces map long-to-short, device_type specific
    ::
        :param device_type: Netmiko device type, increase priority of device_type specific keys
        :param key_lower: True - keys lower-case, False - keys upper-case
        :param value_lower: True - values lower-case, False - values upper-case
        :return: Interfaces map
        :example:
            long_to_short() -> {"Vlan": "V", ...}
            long_to_short(device_type="cisco_ios") -> {"Vlan": "Vlan", ...}
            long_to_short(device_type="cisco_ios", key_lower=True) -> {"vlan": "Vlan", ...}
    """
    data: DStr = short_to_long(device_type=device_type)
    data = {v: k for k, v in data.items()}
    data_ = _overlapped(data)
    data.update(data_)

    if device_type == "cisco_asr":
        data["Tunnel"] = "ti"
    if device_type == "hp_comware":
        data["TenGigabitEthernet"] = "XGE"

    data = _lower(data, key_lower, value_lower)
    return data


def long_to_long(device_type: str = "", key_lower: bool = False, value_lower: bool = False) -> DStr:
    """Returns Interfaces map long-to-long, device_type specific
    ::
        :param device_type: Netmiko device type, increase priority of device_type specific keys
        :param key_lower: True - keys lower-case, False - keys upper-case
        :param value_lower: True - values lower-case, False - values upper-case
        :return: Interfaces map
    """
    data: DStr = short_to_long(device_type=device_type)
    data = {v: k for k, v in data.items()}
    data = {k: k for k in data}
    data = _lower(data, key_lower, value_lower)
    return data


def short_to_long(device_type: str = "",
                  key_lower: bool = False,
                  value_lower: bool = False) -> DStr:
    """Returns Interfaces map short-to-long, device_type specific
    ::
        :param device_type: Netmiko device type, increase priority of device_type specific keys
        :param key_lower: True - keys lower-case, False - keys upper-case
        :param value_lower: True - values lower-case, False - values upper-case
        :return: Interfaces map
        :example:
            short_to_long() -> {"Fa": "FastEthernet", ...}
            short_to_long(key_lower=True) -> {"fa": "FastEthernet", ...}
            short_to_long(value_lower=True) -> {"Fa": "fastethernet", ...}
    """
    if device_type == "cisco_asr":
        data = MAP_OTHER.copy()
        data.update(MAP_CISCO_IOS)
        data.update(MAP_CISCO_ASR)
    elif device_type == "cisco_ios":
        data = MAP_OTHER.copy()
        data.update(MAP_CISCO_IOS)
    elif device_type == "cisco_nxos":
        data = MAP_OTHER.copy()
        data.update(MAP_CISCO_IOS)
        data.update(MAP_CISCO_NXOS)
    elif device_type == "hp_comware":  # h3c
        data = MAP_OTHER.copy()
        data.update(MAP_CISCO_IOS)
        data.update(MAP_HP_COMWARE)
    elif device_type == "hp_procurve":  # h3c
        data = MAP_OTHER.copy()
        data.update(MAP_CISCO_IOS)
        data.update(MAP_HP_PROCURVE)
    else:
        data = MAP_HP_COMWARE.copy()
        data.update(MAP_HP_PROCURVE)
        data.update(MAP_CISCO_ASR)
        data.update(MAP_CISCO_NXOS)
        data.update(MAP_CISCO_IOS)
        data.update(MAP_OTHER)
    data = _lower(data, key_lower, value_lower)
    return data


def short_to_short(device_type: str = "",
                   key_lower: bool = False,
                   value_lower: bool = False) -> DStr:
    """Returns Interfaces map short-to-short, device_type specific
    ::
        :param device_type: Netmiko device type, increase priority of device_type specific keys
        :param key_lower: True - keys lower-case, False - keys upper-case
        :param value_lower: True - values lower-case, False - values upper-case
        :return: Interfaces map
        :example:
            short_to_short(key_lower=True) -> {"fa": "Fa", ...}
            short_to_short(value_lower=True) -> {"Fa": "fa", ...}
    """
    data: DStr = short_to_long(device_type=device_type)
    data = {k: k for k in data}
    intf_map_l2s: DStr = long_to_short(device_type=device_type)
    intf_map_s2l: DStr = short_to_long(device_type=device_type)
    for short_upper, long_upper in intf_map_s2l.items():
        if value := intf_map_l2s.get(long_upper):
            data[short_upper] = value
    data = _lower(data, key_lower, value_lower)
    return data


# ============================== helper ==============================

def _overlapped(data: DStr) -> DStr:
    """Returns Interfaces map long-to-short that overlapped with short key in other device_type

    """
    data_ = {v: k for k, v in MAP_OTHER.items()}
    data_.update({v: k for k, v in MAP_CISCO_IOS.items()})
    data_ = {k: v for k, v in data_.items() if k not in data}
    return data_


def _lower(data: DStr, key_lower: bool, value_lower: bool) -> DStr:
    """Changes key, value to lower-case
    ::
        :param data: Interfaces map data
        :param key_lower: True - keys lower-case, False - keys upper-case
        :param value_lower: True - values lower-case, False - values upper-case
    """
    if key_lower:
        data = {k.lower(): v for k, v in data.items()}
    if value_lower:
        data = {k: v.lower() for k, v in data.items()}
    return data



"""IP Protocols"""

from operator import itemgetter
from typing import Any, Tuple

from netports import helpers as h
from netports.exceptions import NetportsValueError
from netports.ports import snumbers
from netports.range import Range
from netports.static import BRIEF_ALL_I
from netports.types_ import LInt, LStr, DAny, DiAny, SStr, SInt, LTIntStr

IP_NUMBERS: DiAny = {
    0: {"number": 0,
        "name": "hopopt",
        "description": "IPv6 Hop-by-Hop Option, RFC 8200"},
    1: {"number": 1,
        "name": "icmp",
        "description": "Internet Control Message Protocol, RFC 792"},
    2: {"number": 2,
        "name": "igmp",
        "description": "Internet Group Management Protocol, RFC 1112"},
    3: {"number": 3,
        "name": "ggp",
        "description": "Gateway-to-Gateway Protocol, RFC 823"},
    4: {"number": 4,
        "name": "ip-in-ip",
        "description": "IP in IP (encapsulation), RFC 2003"},
    5: {"number": 5,
        "name": "st",
        "description": "Internet Stream Protocol, RFC 1190, RFC 1819"},
    6: {"number": 6,
        "name": "tcp",
        "description": "Transmission Control Protocol, RFC 793"},
    7: {"number": 7,
        "name": "cbt",
        "description": "Core-based trees, RFC 2189"},
    8: {"number": 8,
        "name": "egp",
        "description": "Exterior Gateway Protocol, RFC 888"},
    9: {"number": 9,
        "name": "igp",
        "description": "Interior Gateway Protocol (Cisco IGRP)"},
    10: {"number": 10,
         "name": "bbn-rcc-mon",
         "description": "BBN RCC Monitoring"},
    11: {"number": 11,
         "name": "nvp-ii",
         "description": "Network Voice Protocol, RFC 741"},
    12: {"number": 12,
         "name": "pup",
         "description": "Xerox PUP"},
    13: {"number": 13,
         "name": "argus",
         "description": "ARGUS"},
    14: {"number": 14,
         "name": "emcon",
         "description": "EMCON"},
    15: {"number": 15,
         "name": "xnet",
         "description": "Cross Net Debugger, IEN 158"},
    16: {"number": 16,
         "name": "chaos",
         "description": "Chaos"},
    17: {"number": 17,
         "name": "udp",
         "description": "User Datagram Protocol, RFC 768"},
    18: {"number": 18,
         "name": "mux",
         "description": "Multiplexing, IEN 90"},
    19: {"number": 19,
         "name": "dcn-meas",
         "description": "DCN Measurement Subsystems"},
    20: {"number": 20,
         "name": "hmp",
         "description": "Host Monitoring Protocol, RFC 869"},
    21: {"number": 21,
         "name": "prm",
         "description": "Packet Radio Measurement"},
    22: {"number": 22,
         "name": "xns-idp",
         "description": "XEROX NS IDP"},
    23: {"number": 23,
         "name": "trunk-1",
         "description": "Trunk-1"},
    24: {"number": 24,
         "name": "trunk-2",
         "description": "Trunk-2"},
    25: {"number": 25,
         "name": "leaf-1",
         "description": "Leaf-1"},
    26: {"number": 26,
         "name": "leaf-2",
         "description": "Leaf-2"},
    27: {"number": 27,
         "name": "rdp",
         "description": "Reliable Data Protocol, RFC 908"},
    28: {"number": 28,
         "name": "irtp",
         "description": "Internet Reliable Transaction Protocol, RFC 938"},
    29: {"number": 29,
         "name": "iso-tp4",
         "description": "ISO Transport Protocol Class 4, RFC 905"},
    30: {"number": 30,
         "name": "netblt",
         "description": "Bulk Data Transfer Protocol, RFC 998"},
    31: {"number": 31,
         "name": "mfe-nsp",
         "description": "MFE Network Services Protocol"},
    32: {"number": 32,
         "name": "merit-inp",
         "description": "MERIT Internodal Protocol"},
    33: {"number": 33,
         "name": "dccp",
         "description": "Datagram Congestion Control Protocol, RFC 4340"},
    34: {"number": 34,
         "name": "3pc",
         "description": "Third Party Connect Protocol"},
    35: {"number": 35,
         "name": "idpr",
         "description": "Inter-Domain Policy Routing Protocol, RFC 1479"},
    36: {"number": 36,
         "name": "xtp",
         "description": "Xpress Transport Protocol"},
    37: {"number": 37,
         "name": "ddp",
         "description": "Datagram Delivery Protocol"},
    38: {"number": 38,
         "name": "idpr-cmtp",
         "description": "IDPR Control Message Transport Protocol"},
    39: {"number": 39,
         "name": "tp++",
         "description": "TP++ Transport Protocol"},
    40: {"number": 40,
         "name": "il",
         "description": "IL Transport Protocol"},
    41: {"number": 41,
         "name": "ipv6",
         "description": "IPv6 Encapsulation (6to4 and 6in4), RFC 2473"},
    42: {"number": 42,
         "name": "sdrp",
         "description": "Source Demand Routing Protocol, RFC 1940"},
    43: {"number": 43,
         "name": "ipv6-route",
         "description": "Routing Header for IPv6, RFC 8200"},
    44: {"number": 44,
         "name": "ipv6-frag",
         "description": "Fragment Header for IPv6, RFC 8200"},
    45: {"number": 45,
         "name": "idrp",
         "description": "Inter-Domain Routing Protocol"},
    46: {"number": 46,
         "name": "rsvp",
         "description": "Resource Reservation Protocol, RFC 2205"},
    47: {"number": 47,
         "name": "gre",
         "description": "Generic Routing Encapsulation, RFC 2784, RFC 2890"},
    48: {"number": 48,
         "name": "dsr",
         "description": "Dynamic Source Routing Protocol, RFC 4728"},
    49: {"number": 49,
         "name": "bna",
         "description": "Burroughs Network Architecture"},
    50: {"number": 50,
         "name": "esp",
         "description": "Encapsulating Security Payload, RFC 4303"},
    51: {"number": 51,
         "name": "ah",
         "description": "Authentication Header, RFC 4302"},
    52: {"number": 52,
         "name": "i-nlsp",
         "description": "Integrated Net Layer Security Protocol, TUBA"},
    53: {"number": 53,
         "name": "swipe",
         "description": "SwIPe, RFC 5237"},
    54: {"number": 54,
         "name": "narp",
         "description": "NBMA Address Resolution Protocol, RFC 1735"},
    55: {"number": 55,
         "name": "mobile",
         "description": "IP Mobility (Min Encap), RFC 2004"},
    56: {"number": 56,
         "name": "tlsp",
         "description": "Transport Layer Security Protocol (using Kryptonet key management)"},
    57: {"number": 57,
         "name": "skip",
         "description": "Simple Key-Management for Internet Protocol, RFC 2356"},
    58: {"number": 58,
         "name": "ipv6-icmp",
         "description": "ICMP for IPv6, RFC 4443, RFC 4884"},
    59: {"number": 59,
         "name": "ipv6-nonxt",
         "description": "No Next Header for IPv6, RFC 8200"},
    60: {"number": 60,
         "name": "ipv6-opts",
         "description": "Destination Options for IPv6, RFC 8200"},
    61: {"number": 61,
         "name": "any",
         "description": "host internal protocol"},
    62: {"number": 62,
         "name": "cftp",
         "description": "CFTP"},
    63: {"number": 63,
         "name": "any",
         "description": "local network"},
    64: {"number": 64,
         "name": "sat-expak",
         "description": "SATNET and Backroom EXPAK"},
    65: {"number": 65,
         "name": "kryptolan",
         "description": "Kryptolan"},
    66: {"number": 66,
         "name": "rvd",
         "description": "MIT Remote Virtual Disk Protocol"},
    67: {"number": 67,
         "name": "ippc",
         "description": "Internet Pluribus Packet Core"},
    68: {"number": 68,
         "name": "any",
         "description": "distributed file system"},
    69: {"number": 69,
         "name": "sat-mon",
         "description": "SATNET Monitoring"},
    70: {"number": 70,
         "name": "visa",
         "description": "VISA Protocol"},
    71: {"number": 71,
         "name": "ipcu",
         "description": "Internet Packet Core Utility"},
    72: {"number": 72,
         "name": "cpnx",
         "description": "Computer Protocol Network Executive"},
    73: {"number": 73,
         "name": "cphb",
         "description": "Computer Protocol Heart Beat"},
    74: {"number": 74,
         "name": "wsn",
         "description": "Wang Span Network"},
    75: {"number": 75,
         "name": "pvp",
         "description": "Packet Video Protocol"},
    76: {"number": 76,
         "name": "br-sat-mon",
         "description": "Backroom SATNET Monitoring"},
    77: {"number": 77,
         "name": "sun-nd",
         "description": "SUN ND PROTOCOL-Temporary"},
    78: {"number": 78,
         "name": "wb-mon",
         "description": "WIDEBAND Monitoring"},
    79: {"number": 79,
         "name": "wb-expak",
         "description": "WIDEBAND EXPAK"},
    80: {"number": 80,
         "name": "iso-ip",
         "description": "International Organization for Standardization Internet Protocol"},
    81: {"number": 81,
         "name": "vmtp",
         "description": "Versatile Message Transaction Protocol, RFC 1045"},
    82: {"number": 82,
         "name": "secure-vmtp",
         "description": "Secure Versatile Message Transaction Protocol, RFC 1045"},
    83: {"number": 83,
         "name": "vines",
         "description": "VINES"},
    84: {"number": 84,
         "name": "ttp",
         "description": "TTP"},
    85: {"number": 85,
         "name": "nsfnet-igp",
         "description": "NSFNET-IGP"},
    86: {"number": 86,
         "name": "dgp",
         "description": "Dissimilar Gateway Protocol"},
    87: {"number": 87,
         "name": "tcf",
         "description": "TCF"},
    88: {"number": 88,
         "name": "eigrp",
         "description": "EIGRP, Informational RFC 7868"},
    89: {"number": 89,
         "name": "ospf",
         "description": "Open Shortest Path First, RFC 2328"},
    90: {"number": 90,
         "name": "sprite-rpc",
         "description": "Sprite RPC Protocol"},
    91: {"number": 91,
         "name": "larp",
         "description": "Locus Address Resolution Protocol"},
    92: {"number": 92,
         "name": "mtp",
         "description": "Multicast Transport Protocol"},
    93: {"number": 93,
         "name": "ax.25",
         "description": "AX.25"},
    94: {"number": 94,
         "name": "os",
         "description": "KA9Q NOS compatible IP over IP tunneling"},
    95: {"number": 95,
         "name": "micp",
         "description": "Mobile Internetworking Control Protocol"},
    96: {"number": 96,
         "name": "scc-sp",
         "description": "Semaphore Communications Sec. Pro"},
    97: {"number": 97,
         "name": "etherip",
         "description": "Ethernet-within-IP Encapsulation, RFC 3378"},
    98: {"number": 98,
         "name": "encap",
         "description": "Encapsulation Header, RFC 1241"},
    99: {"number": 99,
         "name": "any",
         "description": "private encryption scheme"},
    100: {"number": 100,
          "name": "gmtp",
          "description": "GMTP"},
    101: {"number": 101,
          "name": "ifmp",
          "description": "Ipsilon Flow Management Protocol"},
    102: {"number": 102,
          "name": "pnni",
          "description": "PNNI over IP"},
    103: {"number": 103,
          "name": "pim",
          "description": "Protocol Independent Multicast"},
    104: {"number": 104,
          "name": "aris",
          "description": "IBM ARIS (Aggregate Route IP Switching) Protocol"},
    105: {"number": 105,
          "name": "scps",
          "description": "SCPS (Space Communications Protocol Standards), SCPS-TP"},
    106: {"number": 106,
          "name": "qnx",
          "description": "QNX"},
    107: {"number": 107,
          "name": "a/n",
          "description": "Active Networks"},
    108: {"number": 108,
          "name": "ipcomp",
          "description": "IP Payload Compression Protocol, RFC 3173"},
    109: {"number": 109,
          "name": "snp",
          "description": "Sitara Networks Protocol"},
    110: {"number": 110,
          "name": "compaq-peer",
          "description": "Compaq Peer Protocol"},
    111: {"number": 111,
          "name": "ipx-in-ip",
          "description": "IPX in IP"},
    112: {"number": 112,
          "name": "vrrp",
          "description": "Virtual Router Redundancy Protocol, "
                         "Common Address Redundancy Protocol (not IANA assigned), 3768"},
    113: {"number": 113,
          "name": "pgm",
          "description": "PGM Reliable Transport Protocol, RFC 3208"},
    114: {"number": 114,
          "name": "any",
          "description": "0-hop protocol"},
    115: {"number": 115,
          "name": "l2tp",
          "description": "Layer Two Tunneling Protocol Version 3, RFC 3931"},
    116: {"number": 116,
          "name": "ddx",
          "description": "D-II Data Exchange (DDX)"},
    117: {"number": 117,
          "name": "iatp",
          "description": "Interactive Agent Transfer Protocol"},
    118: {"number": 118,
          "name": "stp",
          "description": "Schedule Transfer Protocol"},
    119: {"number": 119,
          "name": "srp",
          "description": "SpectraLink Radio Protocol"},
    120: {"number": 120,
          "name": "uti",
          "description": "Universal Transport Interface Protocol"},
    121: {"number": 121,
          "name": "smp",
          "description": "Simple Message Protocol"},
    122: {"number": 122,
          "name": "sm",
          "description": "Simple Multicast Protocol, draft-perlman-simple-multicast-03"},
    123: {"number": 123,
          "name": "ptp",
          "description": "Performance Transparency Protocol"},
    124: {"number": 124,
          "name": "is-is",
          "description": "Intermediate System to Intermediate System (IS-IS) Protocol over IPv4, "
                         "RFC 1142 and RFC 1195"},
    125: {"number": 125,
          "name": "fire",
          "description": "Flexible Intra-AS Routing Environment"},
    126: {"number": 126,
          "name": "crtp",
          "description": "Combat Radio Transport Protocol"},
    127: {"number": 127,
          "name": "crudp",
          "description": "Combat Radio User Datagram"},
    128: {"number": 128,
          "name": "sscopmce",
          "description": "Service-Specific Connection-Oriented Protocol in a Multilink and "
                         "Connectionless Environment, ITU-T Q.2111 (1999)"},
    129: {"number": 129,
          "name": "iplt",
          "description": ""},
    130: {"number": 130,
          "name": "sps",
          "description": "Secure Packet Shield"},
    131: {"number": 131,
          "name": "pipe",
          "description": "Private IP Encapsulation within IP, "
                         "Expired I-D draft-petri-mobileip-pipe-00.txt"},
    132: {"number": 132,
          "name": "sctp",
          "description": "Stream Control Transmission Protocol, "
                         "RFC 4960"},
    133: {"number": 133,
          "name": "fc",
          "description": "Fibre Channel"},
    134: {"number": 134,
          "name": "rsvp-e2e-ignore",
          "description": "Reservation Protocol (RSVP) End-to-End Ignore, RFC 3175"},
    135: {"number": 135,
          "name": "mobility",
          "description": "Header, Mobility Extension Header for IPv6, RFC 6275"},
    136: {"number": 136,
          "name": "udplite",
          "description": "Lightweight User Datagram Protocol, RFC 3828"},
    137: {"number": 137,
          "name": "mpls-in-ip",
          "description": "Multiprotocol Label Switching Encapsulated in IP, RFC 4023, RFC 5332"},
    138: {"number": 138,
          "name": "manet",
          "description": "MANET Protocols, RFC 5498"},
    139: {"number": 139,
          "name": "hip",
          "description": "Host Identity Protocol, RFC 5201"},
    140: {"number": 140,
          "name": "shim6",
          "description": "Site Multihoming by IPv6 Intermediation, RFC 5533"},
    141: {"number": 141,
          "name": "wesp",
          "description": "Wrapped Encapsulating Security Payload, RFC 5840"},
    142: {"number": 142,
          "name": "rohc",
          "description": "Robust Header Compression, RFC 5856"},
    143: {"number": 143,
          "name": "ethernet",
          "description": "IPv6 Segment Routing "
                         "(TEMPORARY - registered 2020-01-31, expired 2021-01-31)"},

}
IP_NAMES_: DAny = {d["name"]: d for i, d in IP_NUMBERS.items()}
IP_ALIASES = {
    "ipinip":
        {"number": 4,
         "name": "ipinip",
         "description": "IP in IP (encapsulation), RFC 2003"},
    "iptm": {"number": 84,
             "name": "iptm",
             "description": "Internet Protocol Traffic Manager"},
}
IP_NAMES = {**IP_NAMES_, **IP_ALIASES}

MIN_NUMBER = 0
MAX_NUMBER = 255
ALL_NUMBERS_L = list(range(MIN_NUMBER, MAX_NUMBER + 1))
ALL_NUMBERS_S = f"{MIN_NUMBER}-{MAX_NUMBER}"


# noinspection PyIncorrectDocstring
def iip(items: Any = "", **kwargs) -> LInt:
    """Sorting IP protocol numbers and removing duplicates
    ::
        :param items: Range of IP protocol numbers, can be unsorted and with duplicates,
            "ip" - Return all IP protocol numbers: [0, 1, ..., 255]
        :type items: str, List[int], List[str]

        :param verbose: True - all protocol numbers in verbose mode: [0, 1, ..., 255],
                        False - all protocol numbers in brief mode: [-1] to save RAM (default)
        :type verbose: bool

        :param strict: True - Raises NetportsValueError, if the protocol is unknown (default)
                       False - Skips unknown protocols

        :param all: True - Return all IP protocol numbers: [0, 1, ..., 255]
        :type all: bool

        :return: *List[int]* of unique sorted IP protocol numbers
        :rtype: List[int]

        :raises NetportsValueError: If IP protocol numbers are outside valid range 0...255

        :example: Converts mix of ip protocol names and numbers to the numbers
            iip("icmp,tcp,7,255") -> [1, 6, 7, 255]

        :example: Converts ip to full range of ip protocol numbers
            iip("ip") -> [0, 1, ... 254, 255]
    """
    if h.is_all(**kwargs):
        if h.is_brief(**kwargs):
            return [BRIEF_ALL_I]
        return ALL_NUMBERS_L.copy()
    if h.is_brief(**kwargs):
        if h.is_brief_in_items(items):
            return [BRIEF_ALL_I]

    items_ = [s.lower() for s in h.split(items)]

    # 0-255
    if ALL_NUMBERS_S in items_:
        if h.is_brief(**kwargs):
            return [BRIEF_ALL_I]
        return ALL_NUMBERS_L.copy()

    # names, numbers
    pairs, invalid = ip_pairs(items=items_, **kwargs)
    if h.is_strict(**kwargs):
        if invalid:
            raise NetportsValueError(f"{invalid=}")
    if not pairs:
        return []
    numbers, names = [list(t) for t in zip(*pairs)]

    if h.is_brief(**kwargs):
        if numbers == ALL_NUMBERS_L:
            return [BRIEF_ALL_I]
    return numbers


# noinspection PyIncorrectDocstring
def sip(items: Any = "", **kwargs) -> str:
    """Soring string IP protocol numbers and removing duplicates
    ::
        :param items: Range of IP protocol numbers, can be unsorted and with duplicates,
            "ip" - mean all numbers in range 0...255
        :type items: str, List[int], List[str]

        :param verbose: True - all protocol numbers in verbose mode: [0, 1, ..., 255],
                        False - all protocol numbers in brief mode: [-1], to save RAM (default)
        :type verbose: bool

        :param strict: True - Raises NetportsValueError, if the protocol is unknown (default),
                       False - Skips unknown protocols
        :type strict: bool

        :param all: True - Return all IP protocol numbers: "0-255"
        :type all: bool

        :return: *str* of unique sorted IP protocol numbers
        :rtype: str

        :raises NetportsValueError: If IP protocol numbers are outside valid range 0...255

        :example:
            sip(["icmp", "tcp", "7", 255]) -> "1,6-7,255"
    """
    if h.is_all(**kwargs):
        return ALL_NUMBERS_S
    if h.is_brief(**kwargs):
        if h.is_brief_in_items(items):
            return ALL_NUMBERS_S

    numbers = iip(items, **kwargs)
    if h.is_brief_in_items(numbers):
        return ALL_NUMBERS_S

    return snumbers(numbers)


# noinspection PyIncorrectDocstring
def ip_pairs(items: Any, **kwargs) -> Tuple[LTIntStr, LStr]:
    """Splits items to IP protocol Number, Name and undefined-invalid protocols
    ::
        :param items: Range of IP protocols, can be unsorted and with duplicates
        :type items: str, List[int], List[str]

        :param verbose: True - all protocol numbers in verbose mode: [0, 1, ..., 255],
                        False - all protocol numbers in brief mode: [-1], to save RAM (default)
        :type verbose: bool

        :return: *List[Tuple[int, str]]* Pairs of IP protocol number and name,
                 *List[str]* Undefined protocol names and invalid numbers
        :rtype: Tuple[List[Tuple[int, str]], List[str]]

        :example:
            ip_pairs("tcp,1,typo,256") -> [(1, "icmp"), (6, "tcp")], ["256", "typo"]
            ip_pairs(["tcp", 1, "typo", 256]) -> [(1, "icmp"), (6, "tcp")], ["256", "typo"]
    """
    items_: LStr = [s.lower() for s in h.split(items)]
    if "ip" in items_:
        if h.is_brief(**kwargs):
            return [(BRIEF_ALL_I, "ip")], []
        return ALL_PAIRS.copy(), []
    if h.is_brief(**kwargs):
        if h.is_brief_in_items(items):
            return [(BRIEF_ALL_I, "ip")], []

    numbers, names = _split_numbers_names(items_)
    names, undefined = _split_names_undefined(names)
    numbers, invalid = _split_numbers_invalid(numbers)
    numbers.update({d["number"] for s, d in IP_NAMES.items() if s in names})
    invalid.update(undefined)

    pairs: LTIntStr = []
    for name in sorted(names):
        number = IP_NAMES[name]["number"]
        if number not in numbers:
            continue
        numbers.remove(number)
        pairs.append((number, name))
    for number in sorted(numbers):
        name = IP_NUMBERS.get(number, {}).get("name") or ""
        pairs.append((number, name))
    pairs.sort(key=itemgetter(0))

    if h.is_brief(**kwargs):
        if pairs == ALL_PAIRS:
            pairs = [(BRIEF_ALL_I, "ip")]

    return pairs, sorted(invalid)


# ============================= helpers ==============================

def all_pairs():
    """Return all pairs of IP protocol (number, name)"""
    paris = []
    for number in ALL_NUMBERS_L:
        name = IP_NUMBERS.get(number, {}).get("name") or ""
        paris.append((number, name))
    return paris


def _split_numbers_names(items: LStr) -> Tuple[SInt, SStr]:
    """Splits `items` to the numbers and names
    ::
        :param items: Combo of numbers and names
        :return: numbers, names
    """
    names: SStr = set()
    numbers: SInt = set()
    for item in items:
        try:
            range_o = Range(item)
            numbers.update(range_o.numbers())
        except NetportsValueError:
            names.add(item)
    return numbers, names


def _split_names_undefined(items: SStr) -> Tuple[SStr, SStr]:
    """Splits `items` to the defined names and undefined names
    ::
        :param items: Combo of defined names and undefined names
        :return: defined names, undefined names
    """
    names: SStr = set()
    undefined: SStr = set()
    for item in items:
        if item in IP_NAMES_:
            names.add(item)
            continue
        if item in IP_ALIASES:
            names.add(item)
            continue
        undefined.add(item)
    return names, undefined


def _split_numbers_invalid(items: SInt) -> Tuple[SInt, SStr]:
    """Splits `items` to the valid protocol numbers and invalid numbers
    ::
        :param items: Combo of valid and invalid numbers
        :return: valid numbers, invalid numbers
    """
    numbers: SInt = set()
    invalid: SInt = set()
    for item in items:
        if item in ALL_NUMBERS_L:
            numbers.add(item)
        else:
            invalid.add(item)
    invalid_: SStr = {str(i) for i in invalid}
    return numbers, invalid_


ALL_PAIRS = all_pairs()



"""IntfGM - Interfaces Group Manager"""
from netports.intf import ULIntf, LIntf, Intf, LStr
from netports.range import Range
from netports.types_ import DSStr


class IntfGM:
    """IntfGM - Interfaces Group Manager"""

    def __init__(self, items):
        """IntfGM
        ::
            :param items: List of Interfaces
            :type items: str, List[str], Intf, List[Intf]
        """
        self.items = items

    def __repr__(self):
        name = self.__class__.__name__
        items = len(self.items)
        return f"<{name}: {items=}>"

    # =========================== property ===========================

    @property
    def items(self) -> LIntf:
        """Interface line"""
        return self._items

    @items.setter
    def items(self, items: ULIntf) -> None:
        if not items:
            _items = []
        elif isinstance(items, (str, Intf)):
            _items = [items]
        elif isinstance(items, (list, set, tuple)):
            _items = list(items)
        else:
            raise TypeError(f"{items=} {list} expected")

        items_: LIntf = []
        for item in _items:
            if isinstance(item, Intf):
                items_.append(item)
            elif isinstance(item, str):
                items_.append(Intf(item))
            else:
                raise TypeError(f"{item=} {str} expected")
        self._items = sorted(items_)

    # =========================== methods ===========================

    def ranges(self, fmt: str = "long") -> LStr:
        """Convert interfaces names to shorted range notation
        ::
            :param fmt: Format option
                "cisco" - Cisco compatible format  # todo
                "long"  - Long names: ["interface Ethernet1/1-3"]
                "short" - Short names: ["Eth1/1/1-3"]
            :type: str

            :return: Interfaces range notation
            :rtype: List[str]
        """
        if fmt == "long":
            return self._ranges__long()
        if fmt == "short":
            return self._ranges__short()
        expected = ["long", "short"]
        raise ValueError(f"{fmt=} {expected=}")

    # =========================== helpers ============================

    def _ranges__long(self) -> LStr:
        """Convert interfaces names to long lines
        ::
            :example:
                intfs = ["interface Ethernet1/1", "interface Ethernet1/2", "interface Ethernet1/3"]
                intf_gm = IntfGM(intfs)
                intf_gm.ranges() -> ["interface Ethernet1/1-3"]
        """
        items = sorted(self.items)

        data: DSStr = {}
        for intf in items:
            idx = intf.last_idx()
            part = intf.part_before(idx)
            data[part] = set()

        for intf in items:
            idx = intf.last_idx()
            part_ = intf.part_before(idx)
            id_ = intf.ids[idx]
            data[part_].add(id_)

        ranges = []
        for key, ids in data.items():
            ids_ = [int(s) for s in ids]
            range_o = Range(ids_)
            for item in range_o.items:
                ranges.append(f"{key}{item}")
        return ranges

    def _ranges__short(self) -> LStr:
        """Convert interfaces names to short lines
        ::
            :example:
                intfs = ["interface Ethernet1/1", "interface Ethernet1/2", "interface Ethernet1/3"]
                intf_gm = IntfGM(intfs)
                intf_gm.ranges() -> ["Eth1/1-3"]
        """
        long_intfs: LStr = self._ranges__long()
        short_intfs: LStr = []
        for long_intf in long_intfs:
            long_intf = long_intf.lower()
            intf_o = Intf(long_intf)
            name_short = intf_o.name_short()
            postfix = long_intf.replace(intf_o.line, "")
            short_intf = name_short + postfix
            short_intfs.append(short_intf)
        return short_intfs


# ============================ functions =============================

def intfrange(items: LStr, fmt: str = "long") -> LStr:
    """Convert interfaces names to shorted range notation
    ::
        :param items: List of interfaces
        :type items: List[str]

        :param fmt: Format option:
            "long"  - Long names: ["interface Ethernet1/1-3"]
            "short" - Short names: ["Eth1/1/1-3"]
        :type: str

        :return: Interface ranges
        :rtype: List[str]

        :example:
            ranges(["Ethernet1/1", "Ethernet1/2", "Ethernet1/3"]) -> ["Ethernet1/1-3"]
    """
    intf_gm = IntfGM(items)
    ranges_ = intf_gm.ranges(fmt=fmt)
    return ranges_



"""Static settings"""

DEVICE_TYPES = [
    "cisco_asr",
    "cisco_ios",
    "cisco_nxos",
    "hp_comware",
    "hp_procurve",
]
RANGE_SPLITTER = "-"
RANGE_SPLITTER_HPE = " to "
RANGE_SPLITTERS = [RANGE_SPLITTER, RANGE_SPLITTER_HPE]
SPLITTER = ","
SPLITTER_HPE = " "

BRIEF_ALL_I = -1
BRIEF_ALL_S = "-1"



"""Item, element of *Range*"""

import re
from functools import total_ordering
from typing import List

from netports.exceptions import NetportsValueError


@total_ordering
class Item:
    """Item, element of *Range*"""

    def __init__(self, line: str):
        """Item, element of *Range*
        ::
            :param line: Item in *str* format
            :type line: str

            :example:
                item = Item("1-3")
                item.line == "1-3"
                item.min == 1
                item.max == 3
                item.range == range(1, 3)
        """
        self.line = line

    def __str__(self):
        return self.line

    def __repr__(self):
        return f"{self.__class__.__name__}({self.line!r})"

    def __hash__(self) -> int:
        return hash((self.min, self.max))

    def __eq__(self, other) -> bool:
        """== equality"""
        if self.__class__ == other.__class__:
            if self.__hash__() == other.__hash__():
                return True
        return False

    def __lt__(self, other) -> bool:
        """< less than"""
        if self.__class__ == other.__class__:
            if self.min != other.min:
                return self.min < other.min
            return self.max < other.max
        return self.line < str(other)

    # =========================== property ===========================

    @property
    def line(self):
        """Item in *str* format"""
        return self._line

    @line.setter
    def line(self, line: str) -> None:
        if not re.match(r"\d+(-\d+)?$", line):
            raise NetportsValueError(f"{line=}, expected range")
        items = line.split("-")
        self._min = int(items[0])
        self._max = self._min if len(items) == 1 else int(items[1])
        if self.min > self.max:
            raise NetportsValueError(f"{self.max=} < {self.min=}")
        self._line = line

    @property
    def min(self):
        """First *int* in range"""
        return self._min

    @property
    def max(self):
        """Last *int* in range"""
        return self._max

    @property
    def range(self):
        """Item in *range* format"""
        return range(self.min, self.max + 1)


LItem = List[Item]
